@@ -4,9 +4,7 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import contextlib
-from packaging.version import Version
 import errno
-from glob import glob
 import hashlib
 import itertools
 import json
@@ -15,32 +13,45 @@ import os
 import os.path as osp
 import re
 import shutil
+import stat
 import site
 import subprocess
 import sys
 import tarfile
+import warnings
+from copy import deepcopy
+from glob import glob
+from pathlib import Path
 from tempfile import TemporaryDirectory
 from threading import Event
-from urllib.request import Request, urlopen, urljoin, quote
 from urllib.error import URLError
-import warnings
+from urllib.request import Request, quote, urljoin, urlopen
 
-from jupyter_core.paths import jupyter_config_path, jupyter_path
-from jupyterlab_server.process import which, Process, WatchHelper, list2cmdline
-from jupyterlab_server.config import LabConfig, get_page_config, get_federated_extensions, get_static_page_config, write_page_config
+from jupyter_core.paths import jupyter_config_path
 from jupyter_server.extension.serverextension import GREEN_ENABLED, GREEN_OK, RED_DISABLED, RED_X
-from traitlets import HasTraits, Bool, Dict, Instance, List, Unicode, default
-
-from jupyterlab.semver import Range, gte, lt, lte, gt, make_semver
-from jupyterlab.jlpmapp import YARN_PATH, HERE
-from jupyterlab.coreconfig import _get_default_core_data, CoreConfig
+from jupyterlab_server.config import (LabConfig, get_federated_extensions,
+                                      get_package_url, get_page_config,
+                                      get_static_page_config,
+                                      write_page_config)
+from jupyterlab_server.process import Process, WatchHelper, list2cmdline, which
+from packaging.version import Version
+from traitlets import Bool, Dict, HasTraits, Instance, List, Unicode, default
 
+from jupyterlab.coreconfig import CoreConfig
+from jupyterlab.jlpmapp import HERE, YARN_PATH
+from jupyterlab.semver import Range, gt, gte, lt, lte, make_semver
+from jupyterlab._version import __version__
 
 # The regex for expecting the webpack output.
-WEBPACK_EXPECT = re.compile(r'.*theme-light-extension/style/index.css')
+WEBPACK_EXPECT = re.compile(r'.*theme-light-extension/style/theme.css')
+
+
+# The repo root directory
+REPO_ROOT = osp.abspath(osp.join(HERE, '..'))
+
 
 # The dev mode directory.
-DEV_DIR = osp.abspath(os.path.join(HERE, '..', 'dev_mode'))
+DEV_DIR = osp.join(REPO_ROOT, 'dev_mode')
 
 
 # If we are pinning the package, rename it `pin@<alias>`
@@ -146,7 +157,9 @@ def get_app_dir():
     """
     # Default to the override environment variable.
     if os.environ.get('JUPYTERLAB_DIR'):
-        return osp.abspath(os.environ['JUPYTERLAB_DIR'])
+        # We must resolve the path to get the canonical case of the path for
+        # case-sensitive systems
+        return str(Path(os.environ['JUPYTERLAB_DIR']).resolve())
 
     # Use the default locations for data_files.
     app_dir = pjoin(sys.prefix, 'share', 'jupyter', 'lab')
@@ -164,7 +177,10 @@ def get_app_dir():
           osp.exists(app_dir) and
           osp.exists('/usr/local/share/jupyter/lab')):
         app_dir = '/usr/local/share/jupyter/lab'
-    return osp.abspath(app_dir)
+
+    # We must resolve the path to get the canonical case of the path for
+    # case-sensitive systems
+    return str(Path(app_dir).resolve())
 
 
 def dedupe_yarn(path, logger=None):
@@ -198,8 +214,7 @@ def ensure_node_modules(cwd, logger=None):
     if ret != 0:
         yarn_proc = ProgressProcess(['node', YARN_PATH], cwd=cwd, logger=logger)
         yarn_proc.wait()
-        parent = pjoin(HERE, '..')
-        dedupe_yarn(parent, logger)
+        dedupe_yarn(REPO_ROOT, logger)
 
     return ret != 0
 
@@ -207,13 +222,12 @@ def ensure_node_modules(cwd, logger=None):
 def ensure_dev(logger=None):
     """Ensure that the dev assets are available.
     """
-    parent = pjoin(HERE, '..')
     logger = _ensure_logger(logger)
-    target = pjoin(parent, 'dev_mode', 'static')
+    target = pjoin(DEV_DIR, 'static')
 
     # Determine whether to build.
-    if ensure_node_modules(parent, logger) or not osp.exists(target):
-        yarn_proc = ProgressProcess(['node', YARN_PATH, 'build'], cwd=parent,
+    if ensure_node_modules(REPO_ROOT, logger) or not osp.exists(target):
+        yarn_proc = ProgressProcess(['node', YARN_PATH, 'build'], cwd=REPO_ROOT,
                             logger=logger)
         yarn_proc.wait()
 
@@ -258,11 +272,10 @@ def watch_packages(logger=None):
     -------
     A list of `WatchHelper` objects.
     """
-    parent = pjoin(HERE, '..')
     logger = _ensure_logger(logger)
-    ensure_node_modules(parent, logger)
+    ensure_node_modules(REPO_ROOT, logger)
 
-    ts_dir = osp.abspath(osp.join(HERE, '..', 'packages', 'metapackage'))
+    ts_dir = osp.abspath(osp.join(REPO_ROOT, 'packages', 'metapackage'))
 
     # Run typescript watch and wait for the string indicating it is done.
     ts_regex = r'.* Found 0 errors\. Watching for file changes\.'
@@ -328,6 +341,8 @@ class AppOptions(HasTraits):
 
     registry = Unicode(help="NPM packages registry URL")
 
+    splice_source = Bool(False, help="Splice source packages into app directory.")
+
     @default('logger')
     def _default_logger(self):
         return logging.getLogger('jupyterlab')
@@ -363,10 +378,8 @@ def watch(app_options=None):
 
     Parameters
     ----------
-    app_dir: string, optional
-        The application directory.
-    logger: :class:`~logger.Logger`, optional
-        The logger instance.
+    app_options: :class:`AppOptions`, optional
+        The application options.
 
     Returns
     -------
@@ -375,7 +388,13 @@ def watch(app_options=None):
     app_options = _ensure_options(app_options)
     _node_check(app_options.logger)
     handler = _AppHandler(app_options)
-    return handler.watch()
+
+    if app_options.splice_source:
+        package_procs = watch_packages(app_options.logger)
+    else:
+        package_procs = []
+
+    return package_procs + handler.watch()
 
 
 
@@ -574,7 +593,8 @@ class _AppHandler(object):
         self.app_dir = options.app_dir
         self.sys_dir = get_app_dir() if options.use_sys_dir else self.app_dir
         self.logger = options.logger
-        self.core_data = options.core_config._data
+        # Make a deep copy of the core data so we don't influence the original copy
+        self.core_data = deepcopy(options.core_config._data)
         self.labextensions_path = options.labextensions_path
         self.kill_event = options.kill_event
         self.registry = options.registry
@@ -640,6 +660,11 @@ class _AppHandler(object):
         if not production:
             minimize = False
 
+        # If splicing, make sure the source packages are built
+        if self._options.splice_source:
+            ensure_node_modules(REPO_ROOT, logger=self.logger)
+            self._run(['node', YARN_PATH, 'build:packages'], cwd=REPO_ROOT)
+
         info = ['production' if production else 'development']
         if production:
             info.append('minimized' if minimize else 'not minimized')
@@ -664,11 +689,8 @@ class _AppHandler(object):
 
         # Build the app.
         dedupe_yarn(staging, self.logger)
-        starting_dir = os.getcwd()
-        os.chdir(staging)
         command = f'build:{"prod" if production else "dev"}{":minimize" if minimize else ""}'
-        ret = self._run(['node', YARN_PATH, 'run', command])
-        os.chdir(starting_dir)
+        ret = self._run(['node', YARN_PATH, 'run', command], cwd=staging)
         if ret != 0:
             msg = 'JupyterLab failed to build'
             self.logger.debug(msg)
@@ -779,10 +801,11 @@ class _AppHandler(object):
 
         # Look for mismatched version.
         static_version = old_jlab.get('version', '')
-        core_version = old_jlab['version']
-        if Version(static_version) != Version(core_version):
-            msg = 'Version mismatch: %s (built), %s (current)'
-            return [msg % (static_version, core_version)]
+        if not static_version.endswith('-spliced'):
+            core_version = old_jlab['version']
+            if Version(static_version) != Version(core_version):
+                msg = 'Version mismatch: %s (built), %s (current)'
+                return [msg % (static_version, core_version)]
 
         shadowed_exts = self.info['shadowed_exts']
 
@@ -807,7 +830,10 @@ class _AppHandler(object):
                     messages.append('%s needs to be removed from build' % ext)
 
         # Look for mismatched dependencies
+        src_pkg_dir = pjoin(REPO_ROOT, 'packages')
         for (pkg, dep) in new_deps.items():
+            if old_deps.get(pkg, '').startswith(src_pkg_dir):
+                continue
             if pkg not in old_deps:
                 continue
             # Skip local and linked since we pick them up separately.
@@ -1150,6 +1176,14 @@ class _AppHandler(object):
         if not version:
             version = self.info['core_data']['jupyterlab']['version']
 
+        splice_source = self._options.splice_source
+        if splice_source:
+            self.logger.debug('Splicing dev packages into app directory.')
+            source_dir = DEV_DIR
+            version = __version__ + '-spliced'
+        else:
+            source_dir = pjoin(HERE, 'staging')
+
         # Look for mismatched version.
         pkg_path = pjoin(staging, 'package.json')
 
@@ -1163,8 +1197,11 @@ class _AppHandler(object):
         for fname in ['index.js', 'bootstrap.js', 'publicpath.js',
                       'webpack.config.js',
                       'webpack.prod.config.js',
-                      'webpack.prod.minimize.config.js',
-                      '.yarnrc', 'yarn.js']:
+                      'webpack.prod.minimize.config.js']:
+            target = pjoin(staging, fname)
+            shutil.copy(pjoin(source_dir, fname), target)
+
+        for fname in ['.yarnrc', 'yarn.js']:
             target = pjoin(staging, fname)
             shutil.copy(pjoin(HERE, 'staging', fname), target)
 
@@ -1174,7 +1211,7 @@ class _AppHandler(object):
             _rmtree(templates, self.logger)
 
         try:
-            shutil.copytree(pjoin(HERE, 'staging', 'templates'), templates)
+            shutil.copytree(pjoin(source_dir, 'templates'), templates)
         except shutil.Error as error:
             # `copytree` throws an error if copying to + from NFS even though
             # the copy is successful (see https://bugs.python.org/issue24564
@@ -1220,16 +1257,41 @@ class _AppHandler(object):
 
         # Then get the package template.
         data = self._get_package_template()
+        jlab = data['jupyterlab']
 
         if version:
-            data['jupyterlab']['version'] = version
+            jlab['version'] = version
 
         if name:
-            data['jupyterlab']['name'] = name
+            jlab['name'] = name
 
         if static_url:
-            data['jupyterlab']['staticUrl'] = static_url
+            jlab['staticUrl'] = static_url
 
+        # Handle splicing of packages
+        if splice_source:
+            # Splice workspace tree as linked dependencies
+            for path in glob(pjoin(REPO_ROOT, 'packages', '*', 'package.json')):
+                local_path = osp.dirname(osp.abspath(path))
+                pkg_data = json.loads(Path(path).read_text(encoding='utf-8'))
+                name = pkg_data['name']
+                if name in data['dependencies']:
+                    data['dependencies'][name] = local_path
+                    jlab['linkedPackages'][name] = local_path
+                if name in data['resolutions']:
+                    data['resolutions'][name] = local_path
+
+            # splice the builder as well
+            local_path = osp.abspath(pjoin(REPO_ROOT, 'builder'))
+            data['devDependencies']['@jupyterlab/builder'] = local_path
+            target = osp.join(staging, 'node_modules', '@jupyterlab', 'builder')
+
+            # Remove node_modules so it gets re-populated
+            node_modules = pjoin(staging, 'node_modules')
+            if osp.exists(node_modules):
+                shutil.rmtree(node_modules, ignore_errors=True)
+
+        # Write the package file
         pkg_path = pjoin(staging, 'package.json')
         with open(pkg_path, 'w') as fid:
             json.dump(data, fid, indent=4)
@@ -1245,12 +1307,14 @@ class _AppHandler(object):
                 f.write(template)
         elif not osp.exists(lock_path):
             shutil.copy(lock_template, lock_path)
+            os.chmod(lock_path, stat.S_IWRITE | stat.S_IREAD)
 
     def _get_package_template(self, silent=False):
         """Get the template the for staging package.json file.
         """
         logger = self.logger
-        data = self.info['core_data']
+        # make a deep copy of the data so we don't influence the core data
+        data = deepcopy(self.info['core_data'])
         local = self.info['local_extensions']
         linked = self.info['linked_packages']
         extensions = self.info['extensions']
@@ -1401,13 +1465,7 @@ class _AppHandler(object):
                 alias = filename[len(PIN_PREFIX):-len(".tgz")]
             else:
                 alias = None
-            # homepage, repository  are optional
-            if 'homepage' in data:
-                url = data['homepage']
-            elif 'repository' in data and isinstance(data['repository'], dict):
-                url = data['repository'].get('url', '')
-            else:
-                url = ''
+            url = get_package_url(data)
             extensions[alias or name] = dict(path=path,
                                     filename=osp.basename(path),
                                     url=url,
@@ -1810,7 +1868,7 @@ class _AppHandler(object):
 
             for key in keys:
                 fname = key[0].replace('@', '') + key[1:].replace('@', '-').replace('/', '-') + '.tgz'
-                data = read_package(os.path.join(tempdir, fname))
+                data = read_package(osp.join(tempdir, fname))
                 # Verify that the version is a valid extension.
                 if not _validate_extension(data):
                     # Valid
@@ -1962,10 +2020,10 @@ def _unlink(path, logger):
 def _rmtree_star(path, logger):
     """Remove all files/trees within a dir, logging errors"""
     for filename in os.listdir(path):
-        file_path = os.path.join(path, filename)
-        if os.path.isfile(file_path) or os.path.islink(file_path):
+        file_path = osp.join(path, filename)
+        if osp.isfile(file_path) or osp.islink(file_path):
             _unlink(file_path, logger)
-        elif os.path.isdir(file_path):
+        elif osp.isdir(file_path):
             _rmtree(file_path, logger)
 
 
@@ -2013,10 +2071,10 @@ def _validate_extension(data):
     if mime_extension and mime_extension not in files:
         messages.append('Missing mimeExtension module "%s"' % mime_extension)
 
-    if themePath and not any(f.startswith(themePath) for f in files):
+    if themePath and not any(f.startswith(str(Path(themePath))) for f in files):
         messages.append('themePath is empty: "%s"' % themePath)
 
-    if schemaDir and not any(f.startswith(schemaDir) for f in files):
+    if schemaDir and not any(f.startswith(str(Path(schemaDir))) for f in files):
         messages.append('schemaDir is empty: "%s"' % schemaDir)
 
     return messages
@@ -2045,7 +2103,7 @@ def _get_static_data(app_dir):
     """Get the data for the app static dir.
     """
     target = pjoin(app_dir, 'static', 'package.json')
-    if os.path.exists(target):
+    if osp.exists(target):
         with open(target) as fid:
             return json.load(fid)
     else:
